#include "camera.h"
#include <math.h>

#define ZBUFFER_MAX_DEPTH 0xffffff

CCamera::CCamera()
{
	m_fZBufDepth = (float)ZBUFFER_MAX_DEPTH;
	m_fAspect = 4.f / 3.f;
}

#define GUARDBAND_SIZE  4096.0f

void CCamera::SetupMatrices()
{
	// projection
	float tanfov = tanf(m_fFov);

	float w = 1.f/((GUARDBAND_SIZE * tanfov) / m_fScrW);
	float h = -(m_fAspect/(4.f/3.f)) * w;

	m_vScale.x =  .5f * m_fScrW / ( m_fScrW / GUARDBAND_SIZE );
	m_vScale.y = (.5f * m_fScrH / ( m_fScrH / GUARDBAND_SIZE ));
	m_vScale.z = -0.5f * (float)m_fZBufDepth;
	m_vScale.w = 1.f;
	
	m_vOffset.x = 2048.f;
	m_vOffset.y = 2048.f;
	m_vOffset.z = 0.5f * (float)m_fZBufDepth;
	m_vOffset.w = 0.f;
	
	float a = (m_fZNear + m_fZFar) / (m_fZNear - m_fZFar);
	float b = (-2.f*m_fZNear*m_fZFar)/(m_fZNear-m_fZFar);

	SetColumn(&m_mProj, 0, Vec4(w,0,0,0));
	SetColumn(&m_mProj, 1, Vec4(0,h,0,0));
	SetColumn(&m_mProj, 2, Vec4(0,0,a,1));
	SetColumn(&m_mProj, 3, Vec4(0,0,b,0));

	// view

	LoadIdentity(&m_mView);
	SetColumn(&m_mView, 3, m_vEye);

	m_mViewInverse = ~(m_mView);
	
	m_mViewProj = m_mProj * m_mView;

}

