#ifndef __CAMERA_H__
#define __CAMERA_H__

#include <geommath/geommath.h>


class CCamera
{
public:
	CCamera();	
	inline const Mat44& GetViewMatrix() 		{ return m_mView;}
	inline const Mat44& GetViewInverseMatrix()	{ return m_mViewInverse; } 
	inline const Mat44& GetViewProjMatrix() 	{ return m_mViewProj;}
	inline const Vec4&  GetOffset()				{ return m_vOffset; }
	inline const Vec4&	GetScale()				{ return m_vScale; }
	
	void Set(const Vec3 &vEye,const  Vec3& vTarget)
	{
		(Vec3&)m_vEye = vEye;
		m_vEye.w = 1.f;
		(Vec3&)m_vTarget = vTarget;
		m_vTarget.w = 1.f;

		SetupMatrices();
	};
	
	void SetSizes(const float fScrW, const float fScrH, const float fFov, const float fZNear, const float fZFar)
	{
		m_fScrW 	= fScrW;
		m_fScrH 	= fScrH;
		m_fFov  	= fFov;
		m_fZNear	= fZNear;
		m_fZFar		= fZFar;

		SetupMatrices();
	}
	
private:

	void SetupMatrices();

	Mat44 m_mViewProj;
	Mat44 m_mView;
	Mat44 m_mViewInverse;
	Mat44 m_mProj;

	float m_fAspect;
	float m_fScrW;
	float m_fScrH;
	float m_fZBufDepth;
	float m_fFov;

	Vec4  m_vOffset;
	Vec4  m_vScale;
	
	Vec4  m_vEye;
	Vec4  m_vTarget;

	float m_fZNear, m_fZFar;
};

#endif
