#include "imeshinstance.h"
#include "macro_codes.h"
#include "imesh.h"
#include <math.h>
#include "p2lassert.h"
#include "stitchhack.h"
#include "screen.h"

void IMeshInstance::Draw(Mat44 m, CCamera &cam)
{
	// I kind of regret this bit. In the future
	// allocating a small linear hunk of mem (512 bytes maybe)
	// for every mesh instance seems like a better idea for
	// dma data that requires realtime updating such as a
	// transformation matrix.
	for(int a=0; a<8; a++)
	{
		memcpy(m_pLocToWorld[a], (void *) &m.m_fMat[(a>>1)*4+((a&1)<<1)], 8);
		memcpy(m_pWorldToScreen[a], (void *) &cam.GetViewProjMatrix().m_fMat[(a>>1)*4+((a&1)<<1)], 8);
	}
	
	asm("sync.l");
}

void IMeshInstance::InitDMA_List()
{
	InitChain(m_pDmaVif1, STCYCL(1,1));
	AddWordToChain(m_pDmaVif1, UNPACK_TO_TOP(V4_32, 8, PAD_CHUNK+0));	

	// clear space for, LocalToWorld and WorldToScreen
	for(int a=0; a<4; a++)
	{
		// beware of TTE/NewTag discontinuence
		m_pLocToWorld[a*2+0] = AddWordToChain(m_pDmaVif1, 0);
		AddWordToChain(m_pDmaVif1, 0);
		m_pLocToWorld[a*2+1] = AddWordToChain(m_pDmaVif1, 0);
		AddWordToChain(m_pDmaVif1, 0);
	}
	for(int a=0; a<4; a++)
	{
		// beware of TTE/NewTag discontinuence
		m_pWorldToScreen[a*2+0] = AddWordToChain(m_pDmaVif1, 0);
		AddWordToChain(m_pDmaVif1, 0);
		m_pWorldToScreen[a*2+1] = AddWordToChain(m_pDmaVif1, 0);
		AddWordToChain(m_pDmaVif1, 0);
	}
	// setup info on obj
	AddWordToChain(m_pDmaVif1, UNPACK_TO_TOP(V4_32, 1, PAD_CHUNK+19));
	AddWordToChain(m_pDmaVif1, 0);	// 0 clipping planes
	float fVar = 0;
	AddWordToChain(m_pDmaVif1, *((uint32 *) &fVar));
	fVar = -255.0f / (1000.0f - 20.0f);
	AddWordToChain(m_pDmaVif1, *((uint32 *) &fVar));	// fogval1
	fVar = -1000.0f * fVar;
	AddWordToChain(m_pDmaVif1, *((uint32 *) &fVar));	// fogval2
	
	// Terminate (add a ret_tag)
	Terminate(m_pDmaVif1);	
}



                                      
IMeshInstance::IMeshInstance(IMesh * pMotherMesh, CDmaVif1 * pDmaVif1)
{
	m_pMotherMesh = pMotherMesh;
	m_pDmaVif1 = (CDmaVif1 *) malloc(sizeof(CDmaVif1));
	*m_pDmaVif1 = *pDmaVif1;
	m_pDmaVif1->Reset();
	m_pDmaTags = m_pDmaVif1->GetCurLocation();

	InitDMA_List();

	m_iNumClipPlanes = 0;
}

