#undef DEBUG_DMA

#include <stdio.h>
#include <math.h>
#include <signal.h>
#include "dmavif1.h"
#include "dmagif.h"
#include "screen.h" 
#include <geommath/geommath.h>
#include "camera.h"
#include "p2lassert.h"
#include "time.h"
#include "joypadlib.h"
#include "vu1draw.h"
int loop_again=1;

#define STICK_SENSITIVITY 0.5f
#define POD_SZE 3

#define T_MODE_CLKS_M		(0x03<< 0)
#define T_MODE_GATE_M		(0x01<< 2)
#define T_MODE_GATS_M		(0x01<< 3)
#define T_MODE_GATM_M		(0x03<< 4)
#define T_MODE_ZRET_M		(0x01<< 6)
#define T_MODE_CUE_M		(0x01<< 7)
#define T_MODE_CMPE_M		(0x01<< 8)
#define T_MODE_OVFE_M		(0x01<< 9)
#define T_MODE_EQUF_M		(0x01<<10)
#define T_MODE_OVFF_M		(0x01<<11)

typedef struct 
{
	// 1st quad
	float fUSpeed;
	float fVSpeed;
	float fPosAmplitude;
	float fNorAmplitude;
	// 2nd quad
	float fTime;
	float fPeriod;
	
}tWobbleParams;


tWobbleParams g_WobbleParams = {
									3.6,			// U speed
									0.6,			// V speed
									14.4, 			// Pos amplitude
									0.04,			// Nor amplitude
									0.f,			// time (could be anything)
									1.2				// Period
								};

void UploadATexture(void * pDmaMem);

// this is executed when ctrl-c is pressed, ideally it stops the loop.
// the deinit functions have to be at the end of the main after the loop.
void sig_handle(int sig)
{
	loop_again=0;
}


#include <intmd/intmdloader.h>
#include "c3dsmesh.h"
#include "imesh.h"
#include "imeshinstance.h"


CScreen *cScreen=NULL;
IMesh	myMesh;
IMeshInstance * pMesh;

volatile float g_fTime = 0;

void SyncStitch(CDmaWrap * pDmaWrap)
{
	uint32 uCur = (uint32) pDmaWrap->GetCurLocation();
	assert((uCur&0xf)==0);
	if((uCur&0xff0)==0xff0)
	{
		pDmaWrap->AttachSrcNextTag((void *) (uCur+0x10));
		pDmaWrap->EndPacket();
	}
}

// distribute sps2 specific global pointers to relevant classes
void  DistribGlobPtrs(void * pDescCached, void * pDescUncached)
{
	CDmaWrap::m_pDescCached = (sps2Memory_t *) pDescCached;
	CDmaWrap::m_pDescUncached = (sps2Memory_t *) pDescUncached;
}
int iSPS2Device;

void delete_screen()
{
	joypadlib_shutdown();
	
	if (cScreen!=NULL) delete cScreen;
	cScreen=NULL;
	printf("yay\n");
}

char meshName[256] = {0};
char dataPath[] = "./data/";

inline char *MeshAsc(char* basename)
{
	static char ascPath[256];
	sprintf(ascPath,"%sscene_%s.asc",dataPath,basename);
	return ascPath;
};

inline char *MeshBin(char* basename)
{
	static char ascPath[256];
	sprintf(ascPath,"%sgeomdata_%s.bin",dataPath,basename);
	return ascPath;
};

int main()
{
	FILE* f = fopen("input.dat","rt");

	
	if (f)
	{
		
		fscanf(f,"%s", meshName);
		printf("%s\n", meshName);
		fscanf(f,"%f", &g_WobbleParams.fUSpeed);
		fscanf(f,"%f", &g_WobbleParams.fVSpeed);
		fscanf(f,"%f", &g_WobbleParams.fPosAmplitude);
		fscanf(f,"%f", &g_WobbleParams.fNorAmplitude);
		fscanf(f,"%f", &g_WobbleParams.fPeriod);
		fclose(f);
	}
	else
	{
		strcpy(meshName, "patch");
	}
	

	
	// Open sps2-library and PS2Linux stuff
	iSPS2Device = sps2Init();

	if (iSPS2Device<0) {
	  printf("Can't open SPS2 device.  Is the module loaded?\n");
	  exit(-1);
	}

	// prealloc memory used by the application
	uint32 uAllocDmaMegs = 4;
	sps2Memory_t * pDescCached = sps2Allocate(uAllocDmaMegs*1024*1024, SPS2_MAP_BLOCK_4K | SPS2_MAP_CACHED, iSPS2Device);
	sps2Memory_t * pDescUncached = sps2Remap(pDescCached, SPS2_MAP_UNCACHED, iSPS2Device);
	DistribGlobPtrs(pDescCached, pDescUncached);
	void * pAllDmaMem = pDescCached->pvStart;
	int iVPU1; iVPU1=open("/dev/ps2vpu1", O_RDWR);
	for (int sig=0; sig<128; sig++)	// catch as many signals as possible
		signal(sig, sig_handle);
	
	// Setup Screen
	cScreen = new CScreen;
	atexit(delete_screen);
	cScreen->ScreenSetup();	
	
	// Upload a texture
	UploadATexture(pAllDmaMem);
	
	// Setup a Camera
	float fScrWdth = cScreen->GetWidth();
	float fScrHgt = cScreen->GetHeight();
	//float fDepthZ = cScreen->GetDepth();
	CCamera cCam;
	cCam.SetSizes(fScrWdth, fScrHgt,0.524f,1.f,10000.f);
	cCam.Set(Vec3(0,0,500),Vec3(0,0,0));

	// prelaunch some vif regs
	const uint32 uVifCodeIni[] _ALIGN16 =
	{
		OFFSET((1024-78)>>1),
		BASE(0),
		MSCALL(4),
		NOP
	};
	DPUT_EE_VIF1_FIFO(((uint128 *) uVifCodeIni)[0]);
	
	// Upload VU1Code
	uint32 iLengthCodeInQuads = sizeof(MPGvu1draw) >> 4;;
	void * pDst = (void *) VU1_MICRO_MEM;
	void * pSrc = (void *) MPGvu1draw;
	memcpy(pDst, pSrc, iLengthCodeInQuads*16);
	
	// setup scene
	CSceneCreatorIntmd scene;
	scene.LoadToScene(MeshAsc(meshName), MeshBin(meshName));

	void * pMeshData = (void *) pAllDmaMem;
	CDmaVif1 cDmaVif1(pMeshData, (uAllocDmaMegs*1024*1024)/16);
	C3dsMesh meshtmp;
	meshtmp.SetupMesh(&cDmaVif1, &scene.GetSceneIntmd().GetMesh(0));
	myMesh = meshtmp;

	// cleanup unused data
	scene.CleanUp();
	
	// create application instances
	// intmdloader should really dictate
	// the number of instances
	uint32 uEnd = (uint32) pDescUncached->pvEnd;
	uint32 uOverFlowSize = uEnd - ((uint32) cDmaVif1.GetCurLocation());
	void * pCurLoc = cDmaVif1.GetCurLocation();

	CDmaVif1 cDmaVif1Inst(pCurLoc, uOverFlowSize>>4);
	pMesh = myMesh.CreateInstance(&cDmaVif1Inst);
	pCurLoc = pMesh->GetDmaChain()->GetCurLocation();
	uOverFlowSize = uEnd - ((uint32) pMesh->GetDmaChain()->GetCurLocation());
	
	// setup call chain
	CDmaVif1 cRenderChain(pCurLoc, uOverFlowSize>>4);
		
	// Flush Cache
	sps2FlushCache(iSPS2Device);

	float fAx=0, fAy=0, fAz=0;

	unsigned long buttons;
	unsigned char button_pressures[PS2PAD_NUM_PRESSURE_BUTTONS];
	float left_stick[2],  right_stick[2];
	int active_joypad = 0;
	
	joypadlib_initialize();

	
	float vPos[3] = {0,0,0};

	// fix the wobble
	int iUSpeedFactor = (int)(g_WobbleParams.fUSpeed / g_WobbleParams.fPeriod);
	g_WobbleParams.fUSpeed = g_WobbleParams.fPeriod * (float) iUSpeedFactor;
	
	int iVSpeedFactor = (int)(g_WobbleParams.fVSpeed / g_WobbleParams.fPeriod);
	g_WobbleParams.fVSpeed = g_WobbleParams.fPeriod * (float) iVSpeedFactor;

	int iTex = 1;
	int iWire = 0;

	int bLastSelect = 0;
	int bLastStart  = 0;
	 
	while(loop_again)
	{
		const float PI = (3.141592f);
	
		joypadlib_update();		
		
		buttons = joypadlib_get_joypad_buttons(active_joypad, button_pressures, PRESSURE_TYPE_BYTE);
		joypadlib_get_joypad_sticks(active_joypad, left_stick, right_stick);

		if (buttons & PS2PAD_BUTTON_L1) 
		{
			vPos[2] += 1.f+button_pressures[PS2PAD_PRESSURE_L1];	
		}		

		if (buttons & PS2PAD_BUTTON_L2) 
		{
			vPos[2] -= 1.f+button_pressures[PS2PAD_PRESSURE_L2];	
		}

		if (fabs(left_stick[0]) > STICK_SENSITIVITY)
			vPos[0] += 10.f * (float)left_stick[0];
		if (fabs(left_stick[1]) > STICK_SENSITIVITY)
			vPos[1] -= 10.f * (float)left_stick[1];
		
		if ((buttons & PS2PAD_BUTTON_SELECT) && !bLastSelect)
		{
			iWire = !iWire;
		}

		if ((buttons & PS2PAD_BUTTON_START) && !bLastStart)
		{
			iTex = !iTex;
		}
	
		bLastSelect = (int)(buttons & PS2PAD_BUTTON_SELECT);
		bLastStart = (int)(buttons & PS2PAD_BUTTON_START);
		
		DPUT_GS_BGCOLOR(0xff0000);
		
		if (fabs(right_stick[0]) > STICK_SENSITIVITY)
			fAy += 0.07f * right_stick[0];
		if (fabs(right_stick[1]) > STICK_SENSITIVITY)
			fAx += 0.07f * right_stick[1];
		
		// Reset render chain
		cRenderChain.Reset();

		g_fTime += 1.f/60.f;	// fixed timer (sync-ed by V-Blank)
		
		Mat44 mTrans;
		LoadRotation(&mTrans, fAx, fAy, fAz);
		SetColumn(&mTrans, 3, Vec4(vPos[0],vPos[1],vPos[2], 1) );

		pMesh->Draw(mTrans, cCam);
		float tmpfl[4];

		g_WobbleParams.fTime = g_fTime;
			
		// send params
		SyncStitch(&cRenderChain);
		cRenderChain.AttachSrcCntTag(); 
		cRenderChain.add64(0);
		cRenderChain.add64(UNPACK_TO_TOP(V4_32, 2, POD_SZE + 23) << 32 | STCYCL(1,1));
		tmpfl[0] = g_WobbleParams.fUSpeed;
		tmpfl[1] = g_WobbleParams.fVSpeed;
		tmpfl[2] = g_WobbleParams.fPosAmplitude;
		tmpfl[3] = g_WobbleParams.fNorAmplitude;
		sps2FlushCache(iSPS2Device);
		cRenderChain.add128(*(uint128*)tmpfl);
		tmpfl[0] = g_WobbleParams.fTime;
		tmpfl[1] = (2.f * PI) / g_WobbleParams.fPeriod;
		tmpfl[2] = 0.f;
		tmpfl[3] = 0.f;
		sps2FlushCache(iSPS2Device);
		cRenderChain.add128(*(uint128*)tmpfl);
		cRenderChain.EndPacket();
		
		// send scale & offset
		SyncStitch(&cRenderChain);
		cRenderChain.AttachSrcCntTag(); 
		cRenderChain.add64(0);
		cRenderChain.add64(UNPACK_TO_TOP(V4_32, 2, POD_SZE + 25) << 32 | STCYCL(1,1));
		cRenderChain.add128(*(uint128*)&cCam.GetOffset());
		cRenderChain.add128(*(uint128*)&cCam.GetScale());
		cRenderChain.EndPacket();
			
		// giftag
		int iPrim = (iWire)?2:4;
		SyncStitch(&cRenderChain);
		cRenderChain.AttachSrcCntTag(); 
		cRenderChain.add64(0);
		cRenderChain.add64(UNPACK_TO_TOP(V4_32, 1, POD_SZE + 27) << 32 | STCYCL(1,1));
		cRenderChain.add64(	GIFTAG(0, 0, 1, PRIM_SET(iPrim, 1, iTex, 0, 0, 0, 0, 0, 0), 0, 3));
		cRenderChain.add64(	GIFTAG_REGS4(ST, RGBAQ, XYZF2, 0));
		cRenderChain.EndPacket(); 
			
		SyncStitch(&cRenderChain);
		cRenderChain.AttachSrcCallTag(pMesh->GetDMA_Tags());
		cRenderChain.EndPacket();	
			
		SyncStitch(&cRenderChain);
		cRenderChain.AttachSrcCallTag(myMesh.GetDMA_Tags());	// all geometry
		cRenderChain.EndPacket();

			// closeup global dmachain
		SyncStitch(&cRenderChain);
		cRenderChain.AttachSrcEndTag();
		cRenderChain.EndPacket();
		
		DPUT_GS_BGCOLOR(0x0000ff);
		cRenderChain.Launch(iSPS2Device);

		// spin locking
		cRenderChain.Wait();
		DPUT_GS_BGCOLOR(0x0);
		cScreen->Swap();
	}

	return 0;
}

/************************************************************************************
************************ Texture Hack to prove it works aswell **********************
************************************************************************************/
float Abs(float x) { return (x < 0) ? -x : x; }
void UploadATexture(void * pDmaMem)
{
	// pDmaMem is 16 quads from start of page
	// we're going to build the chain at pDmaMem
	// and then put a dummy texture of 64x16 in 32bit (4096bytes)
	// at the next page. This way we won't have to stitch
	uint32 next_page = (((uint32) pDmaMem)+0x1000)&(~0xfff);	
	for(int y=0; y<16; y++)	// Coder texture times T
		for(int x=0; x<64; x++)
		{
			uint32 col=0xff;

			if(y>=2 && y<5)
			{
				if(x>=10 && x<=(63-10))
				{
					float t = Abs(31.5f-x)/21.5f;
					t = (3-2*t)*t*t;
					col = ((int32) (t*0x5a+0x40));
				}
			}
			else if(y>=5 && y<=13)
			{
				if(x>=26 && x<=(63-26))
				{
					float t = Abs(31.5f-x)/5.5f;
					t = (3-2*t)*t*t;
					col = ((int32) (t*0x5a+0x40));
				}
			}
				
			((uint32 *) next_page)[y*64+x]=(col<<16)|(col<<8)|col;
		}
	
	// Flush Cache
	sps2FlushCache(iSPS2Device);

	// Create DmaChain
	CDmaVif1 dmaVif1(pDmaMem, 240);	// not 256 quads since 16quads were given to screen.cpp

	uint32 uDstPtr = (cScreen->GetFirstFreeGsPage()*2048)/64;//0x1a40;
	uint32 uQSize = (64*16*4)/16;

	dmaVif1.AttachSrcCntTag(DIRECT(6)<<32);
	dmaVif1.add64(GIFTAG(4, 0, 0, 0, 0, 1));	// GIFTAG(NLOOP, EOP, PRE, PRIM, FLG, NREG)
	dmaVif1.add64(GIFTAG_REGS1(AD));
	dmaVif1.add64(BITBLTBUF_SET(0, 0, 0, uDstPtr, 1, 0));		// 0x41a40 00000000L
	dmaVif1.add64(BITBLTBUF);
	dmaVif1.add64(TRXPOS_SET(0, 0, 0, 0, 0));
	dmaVif1.add64(TRXPOS);
	dmaVif1.add64(TRXREG_SET(64, 16));
	dmaVif1.add64(TRXREG);
	dmaVif1.add64(TRXDIR_SET(0));
	dmaVif1.add64(TRXDIR);
	dmaVif1.add64(GIFTAG(uQSize, 0, 0, 0, 2, 0));	// GIFTAG(NLOOP, EOP, PRE, PRIM, FLG, NREG)
	dmaVif1.add64(0);
	dmaVif1.EndPacket();

	//for(int page=0; page<((64*16*4)/4096); page++)
		//dmaVif1.AttachSrcRefTag(((char *) next_page)+page*4096, 4096/16, DIRECT((4096/16))<<32);
	dmaVif1.AttachSrcRefTag(((char *) next_page), 4096/16, DIRECT((4096/16))<<32);

	uint64 uNloop =10;

	dmaVif1.AttachSrcEndTag(DIRECT(uNloop+1)<<32);
	dmaVif1.add64(GIFTAG(uNloop, 1, 0, 0, 0, 1));	// GIFTAG(NLOOP, EOP, PRE, PRIM, FLG, NREG)
	dmaVif1.add64(GIFTAG_REGS1(AD));
	dmaVif1.add64(TEST_SET(0, 0, 0, 0, 0, 0, 1, 3));	// 0x70000
	dmaVif1.add64(TEST_1);
	dmaVif1.add64(PABE_SET(0));
	dmaVif1.add64(PABE);
	dmaVif1.add64(TEXA_SET(0,0,0));//x7f, 1, 0x81));	//     0x810000807fL
	dmaVif1.add64(TEXA);
	dmaVif1.add64(FBA_SET(0));
	dmaVif1.add64(FBA_1);
	dmaVif1.add64(0);		// NOP data always 0
	dmaVif1.add64(0xf);	// NOP instruction
	dmaVif1.add64(TEX1_SET(0, 0, 3, 0, 0, 0, 0));		// 0x60
	dmaVif1.add64(TEX1_1);
	dmaVif1.add64(TEX0_SET(uDstPtr, 1, 0, 6, 4, 1, 0, 0, 0, 0, 0, 0));	// 0x620011a40L
	dmaVif1.add64(TEX0_1);
	dmaVif1.add64(CLAMP_SET(0,0,0,0,0,0));
	dmaVif1.add64(CLAMP_1);
	dmaVif1.add64(COLCLAMP_SET(0x1));
	dmaVif1.add64(COLCLAMP);
	dmaVif1.add64(0);
	dmaVif1.add64(TEXFLUSH);
	dmaVif1.EndPacket();
	dmaVif1.Launch();
	dmaVif1.Wait();
	printf("TextureDone\n");
}

